package org.jvnet.jaxbcommons.cloneable.addon;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jvnet.jaxbcommons.addon.AbstractParameterizableCodeAugmenter;
import org.jvnet.jaxbcommons.addon.generator.MethodStrategy;
import org.jvnet.jaxbcommons.cloneable.addon.generator.CloneableStrategy;
import org.jvnet.jaxbcommons.cloneable.addon.tests.RoundtripTest;
import org.jvnet.jaxbcommons.util.GeneratorContextUtils;
import org.xml.sax.ErrorHandler;

import com.sun.tools.xjc.Options;
import com.sun.tools.xjc.generator.ClassContext;
import com.sun.tools.xjc.generator.GeneratorContext;
import com.sun.tools.xjc.grammar.AnnotatedGrammar;
import com.sun.tools.xjc.grammar.ClassItem;

/**
 * Add-on for {@link Object#equals(Object)} and {@link Object#hashCode()}
 * methods generation.
 * 
 * @author Aleksei Valikov
 */
public class AddOn extends AbstractParameterizableCodeAugmenter {

	/**
	 * Logger.
	 */
	protected Log logger = LogFactory.getLog(getClass());

	public String getOptionName() {
		return "Xcloneable";
	}

	public String getUsage() {
		return "  -Xcloneable                  :  makes generated classes cloneable";
	}

	/**
	 * Augments the code.
	 * 
	 * @param grammar
	 *            annotated grammar.
	 * @param generatorContext
	 *            generator context.
	 * @param options
	 *            options.
	 * @param errorHandler
	 *            error handler.
	 * @return If the add-on executes successfully, return true. If it detects
	 *         some errors but those are reported and recovered gracefully,
	 *         return false.
	 */
	public boolean run(final AnnotatedGrammar grammar,
			final GeneratorContext generatorContext, final Options options,
			final ErrorHandler errorHandler) {

		logger.debug("Starting code augmentation.");

		// Process class items
		final ClassItem[] classItems = grammar.getClasses();
		logger.debug("Processing [" + classItems.length + "] class items.");
		for (int index = 0; index < classItems.length; index++) {
			final ClassItem classItem = classItems[index];
			final ClassContext classContext = generatorContext
					.getClassContext(classItem);
			logger.debug("Processing [" + classItem.name + "].");
			getCloneableStrategy().generate(classContext, null,
					classContext.implClass);
		}
		logger.debug("Finished processing class items.");
		logger.debug("Finished code augmentation.");

		final String rt = getRt();
		if (rt != null) {
			logger.debug("Generating roundtrip test [" + rt + "].");
			GeneratorContextUtils.generateContextPathAwareClass(grammar,
					generatorContext, rt, RoundtripTest.class);
			logger.debug("Finished generating roundtrip test.");
		}

		return true;
	}

	protected MethodStrategy cloneableStrategy = new CloneableStrategy();

	public MethodStrategy getCloneableStrategy() {
		return cloneableStrategy;
	}

	public void setStrategy(MethodStrategy cloneableStrategy) {
		this.cloneableStrategy = cloneableStrategy;
	}

	private String rt = null;

	public String getRt() {
		return rt;
	}

	public void setRt(String roundtripTestClassName) {
		this.rt = roundtripTestClassName;
	}
}
