package org.jvnet.jaxbcommons.copyable.addon;

import org.jvnet.jaxbcommons.addon.AbstractParameterizableCodeAugmenter;
import org.jvnet.jaxbcommons.lang.Cloneable;
import org.jvnet.jaxbcommons.lang.CopyTo;
import org.jvnet.jaxbcommons.lang.Copyable;
import org.jvnet.jaxbcommons.lang.builder.CopyBuilder;
import org.jvnet.jaxbcommons.lang.builder.JAXBCopyBuilder;
import org.jvnet.jaxbcommons.util.AccessorUtils;
import org.jvnet.jaxbcommons.util.ClassUtils;
import org.jvnet.jaxbcommons.util.FieldUtils;
import org.xml.sax.ErrorHandler;

import com.sun.codemodel.JBlock;
import com.sun.codemodel.JCodeModel;
import com.sun.codemodel.JDefinedClass;
import com.sun.codemodel.JExpr;
import com.sun.codemodel.JMethod;
import com.sun.codemodel.JMod;
import com.sun.codemodel.JOp;
import com.sun.codemodel.JType;
import com.sun.codemodel.JVar;
import com.sun.tools.xjc.Options;
import com.sun.tools.xjc.generator.ClassContext;
import com.sun.tools.xjc.generator.GeneratorContext;
import com.sun.tools.xjc.generator.field.FieldRenderer;
import com.sun.tools.xjc.grammar.AnnotatedGrammar;
import com.sun.tools.xjc.grammar.ClassItem;
import com.sun.tools.xjc.grammar.FieldItem;
import com.sun.tools.xjc.grammar.FieldUse;

public class AddOn extends AbstractParameterizableCodeAugmenter {

  public String getOptionName() {
    return "Xcopyable";
  }

  public String getUsage() {
    return "TBD";
  }

  private Class copyBuilderClass = JAXBCopyBuilder.class;

  public void setCopyBuilderClass(final Class copyBuilderClass) {
    this.copyBuilderClass = copyBuilderClass;
  }

  public Class getCopyBuilderClass() {
    return copyBuilderClass;
  }

  public boolean run(
      final AnnotatedGrammar grammar,
      final GeneratorContext generatorContext,
      final Options options,
      final ErrorHandler errorHandler) {

    logger.debug("Starting code augmentation.");

    // Process class items
    final ClassItem[] classItems = grammar.getClasses();
    logger.debug("Processing [" + classItems.length + "] class items.");
    for (int index = 0; index < classItems.length; index++) {
      final ClassItem classItem = classItems[index];
      final ClassContext classContext = generatorContext.getClassContext(classItem);
      logger.debug("Processing [" + classItem.name + "].");
      processClassContext(classContext);
    }
    logger.debug("Finished processing class items.");
    logger.debug("Finished code augmentation.");
    return true;

  }

  protected void processClassContext(ClassContext classContext) {

    final JMethod copyTo$copyTo = generateCopyTo$CopyTo(classContext);
    final JMethod copyable$copyTo = generateCopyable$CopyTo(classContext);
    final JMethod cloneable$clone = generateCloneable$Clone(classContext);
    ClassUtils._implements(classContext.ref, classContext.ref.owner().ref(CopyTo.class));

  }

  protected JMethod generateCopyable$CopyTo(final ClassContext classOutline) {
    final JDefinedClass theClass = classOutline.implClass;
    //    ClassUtils._implements(theClass, theClass.owner().ref(Copyable.class));
    //    ClassUtils._implements(classOutline.ref, theClass.owner().ref(Copyable.class));

    final JMethod copyable$copyTo = theClass.method(
        JMod.PUBLIC,
        theClass.owner().ref(Object.class),
        "copyTo");
    {
      final JVar target = copyable$copyTo.param(Object.class, "target");
      final JBlock body = copyable$copyTo.body();

      final JVar copyBuilder = body.decl(
          JMod.FINAL,
          theClass.owner().ref(CopyBuilder.class),
          "copyBuilder",
          JExpr._new(theClass.owner().ref(getCopyBuilderClass())));

      body._return(JExpr.invoke("copyTo").arg(target).arg(copyBuilder));
    }
    return copyable$copyTo;
  }

  protected JMethod generateCloneable$Clone(final ClassContext classOutline) {
    final JDefinedClass theClass = classOutline.implClass;
    //    ClassUtils._implements(theClass, theClass.owner().ref(Cloneable.class));
    //    ClassUtils._implements(classOutline.ref, theClass.owner().ref(Cloneable.class));

    final JMethod cloneable$Clone = theClass.method(
        JMod.PUBLIC,
        theClass.owner().ref(Object.class),
        "clone");
    {
      final JBlock body = cloneable$Clone.body();

      final JVar target = body.decl(classOutline.ref, "target", JExpr._new(theClass));

      body._return(JExpr.invoke("copyTo").arg(target));
    }
    return cloneable$Clone;
  }

  protected JMethod generateCopyTo$CopyTo(ClassContext classOutline) {
    
    final JDefinedClass theClass = classOutline.implClass;
    final JCodeModel codeModel = theClass.owner();

    final JMethod copyTo = theClass.method(
        JMod.PUBLIC,
        theClass.owner().ref(Object.class),
        "copyTo");
    {
      final JVar target = copyTo.param(Object.class, "target");
      final JVar copyBuilder = copyTo.param(CopyBuilder.class, "copyBuilder");

      final JBlock body = copyTo.body();

      final JVar copy = body.decl(JMod.FINAL, theClass, "copy",

      JOp.cond(JOp.eq(target, JExpr._null()), JExpr._new(theClass), JExpr.cast(theClass, target)));

      if (classOutline.target.getSuperClass() != null) {
        body.invoke(JExpr._super(), "copyTo").arg(copy).arg(copyBuilder);
      }

      final FieldUse[] fieldUses = classOutline.target.getDeclaredFieldUses();

      for (int index = 0; index < fieldUses.length; index++) {
        final FieldUse fieldUse = fieldUses[index];
        final FieldItem fieldItem = FieldUtils.getFieldItem(fieldUse);

        if (!FieldUtils.isConstantField(classOutline, fieldItem)) {

          final JBlock block = body.block();
          
          JMethod get = AccessorUtils.get(classOutline, fieldItem);
          final JType type = get.type();

          final JVar sourceField = block.decl(type, "source" + fieldUse.name);

          AccessorUtils.get(block, sourceField, JExpr._this(), classOutline, fieldItem);

          final JVar copyField = block.decl(type, "copy" + fieldItem.name,

          JExpr.cast(type, JExpr.invoke(copyBuilder, "copy").arg(sourceField)));

          //          JMethod set = AccessorUtils.set(classOutline, fieldItem);
          //          set = AccessorUtils.set(classOutline, fieldItem);
          //          block.invoke(copy, set).arg(copyField);
          AccessorUtils.set(block, copyField, copy, classOutline, fieldItem);
        }
      }
      body._return(copy);
      return copyTo;
    }
  }
}
