package org.jvnet.jaxbcommons.util;

import java.util.List;

import com.sun.codemodel.JBlock;
import com.sun.codemodel.JCodeModel;
import com.sun.codemodel.JDefinedClass;
import com.sun.codemodel.JExpression;
import com.sun.codemodel.JMethod;
import com.sun.codemodel.JType;
import com.sun.codemodel.JVar;
import com.sun.tools.xjc.generator.ClassContext;
import com.sun.tools.xjc.grammar.FieldItem;

/**
 * Accessor utilities.
 * 
 * @author valikov
 */
public class AccessorUtils {

  /**
   * Hidden constructor.
   */
  private AccessorUtils() {
  }

  public static void get(
      JBlock block,
      JVar var,
      JExpression target,
      final ClassContext classContext,
      final FieldItem fieldItem) {
    final JMethod method = get(classContext, fieldItem);
    block.assign(var, target.invoke(method));
  }
  
  public static void set(
      JBlock block,
      JExpression value,
      JExpression target,
      final ClassContext classContext,
      final FieldItem fieldItem) {
    final JCodeModel codeModel = classContext.ref.owner();
    final JType type = fieldItem.getType(codeModel);
    
    final JMethod set = set(classContext, fieldItem);
    if (set == null)
    {
      final JMethod get = get(classContext, fieldItem);
      block.add(target.invoke(get).invoke("clear"));
      block.add(target.invoke(get).invoke("addAll").arg(value));
    }
    else
    {
      block.invoke(target, set).arg(value);
    }
  }
  

  public static JMethod get(final ClassContext classContext, final FieldItem fieldItem) {
    final JCodeModel codeModel = classContext.ref.owner();
    final JType type = fieldItem.getType(codeModel);
    final String getterName = (codeModel.BOOLEAN.equals(type) ? "is" : "get") + fieldItem.name;
    return CodeModelUtils.getMethod(classContext, getterName);
  }

  public static JMethod set(final ClassContext classContext, final FieldItem fieldItem) {
    final String setterName = "set" + fieldItem.name;
    return CodeModelUtils.getMethod(classContext, setterName);
  }

  public static JMethod unset(final ClassContext classContext, final FieldItem fieldItem) {
    final String unsetterName = "unset" + fieldItem.name;
    final JMethod method = CodeModelUtils.getMethod(classContext, unsetterName);
    if (method == null)
      throw new AssertionError("No unsetter [" + unsetterName + "] found.");
    return method;
  }

  public static JMethod isSet(final ClassContext classContext, final FieldItem fieldItem) {
    final String isSetterName = "isSet" + fieldItem.name;
    return CodeModelUtils.getMethod(classContext, isSetterName);
  }
}