package org.jvnet.jaxbcommons.util;

import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import com.sun.codemodel.JClass;
import com.sun.codemodel.JDefinedClass;
import com.sun.tools.xjc.generator.ClassContext;
import com.sun.tools.xjc.grammar.ClassItem;

public class ClassUtils {
  
  public static String getFQCN(final ClassContext theClass, boolean dotted) {
    return CodeModelUtils.getFQCN(theClass.ref, dotted);
  }

  public static String getLCN(final ClassContext theClass, boolean dotted) {
    return CodeModelUtils.getLCN(theClass.ref, dotted);
  }
  
  public static String getFQCN(final ClassItem theClass, boolean dotted) {
    return CodeModelUtils.getFQCN(theClass.getTypeAsDefined(), dotted);
  }

  public static String getLCN(final ClassItem theClass, boolean dotted) {
    return CodeModelUtils.getLCN(theClass.getTypeAsDefined(), dotted);
  }
  

  public static String getClassName(final ClassContext classContext) {
    return getClassName(classContext.target);
  }
  
  public static String getPackagedClassName(final ClassContext classContext) {
    return CodeModelUtils.getPackagedClassName(classContext.target.getTypeAsDefined());
  }
  

  public static String getClassName(final ClassItem classItem) {
    return CodeModelUtils.getClassName(classItem.getTypeAsDefined());
  }
  

  public static void _implements(JDefinedClass theClass, JClass theInterface) {
    if (!isImplementing(theClass, theInterface))
      theClass._implements(theInterface);
  }

  public static boolean isImplementing(JDefinedClass theClass, JClass theInterface) {
    for (Iterator iterator = theClass._implements(); iterator.hasNext();) {
      final JClass implementedInterface = (JClass) iterator.next();
      if (theInterface.equals(implementedInterface)) {
        return true;
      }
    }
    return false;
  }
  
  public static List getAncestors(ClassContext classContext)
  {
	  final ClassItem classItem = classContext.target;
	  if (classItem.getSuperClass() == null)
		  return Collections.EMPTY_LIST;
	  else
	  {
		  final ClassItem superClassItem = classItem.getSuperClass();
		  final ClassContext superClassContext = classContext.parent.getClassContext(superClassItem);
		  return getAncestorsAndSelf(superClassContext);
	  }
  }

  public static List getAncestorsAndSelf(ClassContext classContext)
  {
	  final ClassItem classItem = classContext.target;
	  if (classItem.getSuperClass() == null)
		  return Collections.singletonList(classContext);
	  else
	  {
		  final List superClassContexts = new LinkedList();
		  superClassContexts.add(classContext);
		  final ClassItem superClassItem = classItem.getSuperClass();
		  final ClassContext superClassContext = classContext.parent.getClassContext(superClassItem);
		  superClassContexts.addAll(getAncestorsAndSelf(superClassContext));
		  return superClassContexts;
	  }
  }
  
}
